#######################################################################
#
# pbsclient component
#
# Initialise an PBS client
#
# David Groep <davidg@nikhef.nl>
# Version 0.1.3 : 30/12/04 14:44
#
# Copyright (c) 2004 Thorsten Kleinwort, CERN and EU DataGrid.
# For license conditions see http://www.eu-datagrid.org/license.html
#
#
#######################################################################

package NCM::Component::pbsclient;

#
# a few standard statements, mandatory for all components
#

use strict;
use LC::Check;
use NCM::Check;
use NCM::Component;
use vars qw(@ISA $EC);
@ISA = qw(NCM::Component);
$EC=LC::Exception::Context->new->will_store_all;

my ($pbsmomconf)="/var/spool/pbs/mom_priv/config";
my ($pbsdir)="/var/spool/pbs";
my ($pbsmomdir)="/var/spool/pbs/mom_priv";
my ($pbsinitscript)="/etc/init.d/pbs";
my ($masterlist, $restrictedlist, $logevent, $tmpdir, %resources);
my ($resources, $cpuf, $wallf, $idealload, $maxload);
my (%usecp, $prologalarm, $behaviour, $nodecheckscript);
my ($nodecheckinterval);

sub initpaths {
  my ($config) = (@_);

  # where should the config file be written? Will default
  if ( $config->elementExists("/software/components/pbsclient/configPath")){
    $pbsmomconf = 
      $config->getValue("/software/components/pbsclient/configPath");
  }
  # name of the init script to restart pbs if config changed
  if ( $config->elementExists("/software/components/pbsclient/initScriptPath")){
    $pbsinitscript = 
      $config->getValue("/software/components/pbsclient/initScriptPath");
  }
 
  # make sure pbs mom directory exists and it properly writable
  ($pbsmomdir=$pbsmomconf) =~ s/\/[^\/]+$//; # implements dirname()
  ($pbsdir=$pbsmomdir) =~ s/\/[^\/]+$//; # implements dirname() again

}

sub retrieve {
  my ($self,$config,$path,$isreq,$scref,$default) = @_;
  my ($tmpval);

  $self->debug(1,"Retrieving resource in $path");

  if ( $config->elementExists($path) ) {
    $tmpval = $config->getValue($path);
    $$scref=$tmpval;
    $self->debug(1,"Set value of ref ".$scref." to $tmpval");
  } else {
    if ($isreq) {
      $self->Error("Cannot obtain required value from $path");
    } else {
      $$scref=$default;

      $default or $default="<undef>";
      $self->debug(1,"Set value of ref ".$scref." to default $default");
    }
  }
}

##########################################################################
sub Configure {
##########################################################################
  my ($self,$config)=@_;
  my ($n) = 0;

  $self->info("Configuring PBS");

  &initpaths($config);

  $self->debug(2,"PBSMOMDIR $pbsmomdir; PBSDIR $pbsdir");

  # the masterlist will fill the $clienthost directive
  $masterlist="";
  $n=0;
  while ($config->elementExists(
            "/software/components/pbsclient/masters/".$n)) {
    ($masterlist ne "") and $masterlist.=" ";
    $masterlist .=
      $config->getValue("/software/components/pbsclient/masters/".$n);
    $n++;
  }
  if ($masterlist eq ""){
    $self->error("Empty master list");
    return;
  }

  # the restrictedlist will fill the $restricted directive
  $restrictedlist="";
  $n=0;
  while ($config->elementExists(
            "/software/components/pbsclient/restricted/".$n)) {
    ($restrictedlist ne "") and $restrictedlist.=" ";
    $restrictedlist .=
      $config->getValue("/software/components/pbsclient/restricted/".$n);
    $n++;
  }
  if ($restrictedlist eq ""){
    $self->info("No additional hosts added to the restricted list");
  }

  # additional resources (those without a $) that end up in the config

  $n=0;
  while (
      $config->elementExists("/software/components/pbsclient/resources/$n") ) {
    my $resource = 
      $config->getValue("/software/components/pbsclient/resources/$n");
    my ($a,$v)=split(/:/,$resource,2);
    $resources{$a}=$v;
    $n++;
  } 
  $n or $self->info("No additional resources defined");

  # can we add Torque-specific stuff (e.g. location of the server name)?
  $self->retrieve($config,
                  "/software/components/pbsclient/behaviour",0,
                  \$behaviour,"OpenPBS");

  # verbosity of the pbs mom
  $self->retrieve($config,
                  "/software/components/pbsclient/logEvent",0,
                  \$logevent,undef);

  # tmpdir (from the transient_tmpdir patch)
  $self->retrieve($config,
                  "/software/components/pbsclient/tmpdir",0,
                  \$tmpdir,undef);
  
  # idealload
  $self->retrieve($config,
                  "/software/components/pbsclient/idealLoad",0,
                  \$idealload,undef);
 
  # maxload
  $self->retrieve($config,
                  "/software/components/pbsclient/maxLoad",0,
                  \$maxload,undef);

  # cpuFactor
  $self->retrieve($config,
                  "/software/components/pbsclient/cpuFactor",0,
                  \$cpuf,undef);

  # wallFactor
  $self->retrieve($config,
                  "/software/components/pbsclient/wallFactor",0,
                  \$wallf,undef);

  # prologAlarmSec
  $self->retrieve($config,
                  "/software/components/pbsclient/prologAlarmSec",0,
                  \$prologalarm,undef);

  # nodeCheckScriptPath
  $self->retrieve($config,
                  "/software/components/pbsclient/nodeCheckScriptPath",0,
                  \$nodecheckscript,undef);

  # nodeCheckIntervalSec
  $self->retrieve($config,
                  "/software/components/pbsclient/nodeCheckIntervalSec",0,
                  \$nodecheckinterval,undef);

  # additional usecp directives, list of lists(2)
  $n=0;
  while (
      $config->elementExists("/software/components/pbsclient/directPaths/$n")){
    # do both elements exist?
    $self->Error("Cannot find mountloc for directPath $n") unless
      $config->elementExists("/software/components/pbsclient/directPaths/$n/locations");
    $self->Error("Cannot find location for directPath $n") unless
      $config->elementExists("/software/components/pbsclient/directPaths/$n/path");

    my $locations = 
      $config->getValue("/software/components/pbsclient/directPaths/$n/locations");
    my $path = 
      $config->getValue("/software/components/pbsclient/directPaths/$n/path");

    $usecp{$locations}=$path;
    $self->debug(1,"Adding direct path $n: \$usecp $locations $path");
    $n++;
  } 



  LC::Check::directory("$pbsmomdir");

  -e "$pbsmomconf" or ( open PBSMOMCONF,">$pbsmomconf" and close PBSMOMCONF );
  -f "$pbsmomconf" or $self->Error("$pbsmomconf exists but is not a file");

  LC::Check::status("$pbsmomconf",
	owner => "root",
	group => "root",
        mode  => 0640 );

  $self->Info("Modifying file $pbsmomconf");

  my $changes=0;

  my $cnflines="";
  my $cnfre="";
  my $cnfgoodre="";

  # create the line(s) with the $clienthost directives from the master list
  foreach ( split(/\s+/,$masterlist) ) { 
    $cnflines.='$clienthost '.$_."\n";
    $cnfgoodre.='\\$clienthost '.$_."\n";
    $cnfre.='\\$clienthost\\s+.*'."\n";
  }
  chomp($cnflines);
  chomp($cnfre);
  chomp($cnfgoodre);
  $cnfgoodre=~s+\.+\\\.+g;

  $changes+=NCM::Check::lines($pbsmomconf,
        backup => ".old",
	linere => $cnfre,
	goodre => $cnfgoodre,
	good   => $cnflines,
        keep   => "first",
        add    => "first"
        );

  $tmpdir and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$tmpdir\\s+.*',
        goodre => '^\\$tmpdir\\s+'.$tmpdir,
        good   => "\$tmpdir $tmpdir",
        keep   => "first",
        add    => "last"
        );

  $logevent and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$logevent\\s+.*',
        goodre => '^\\$logevent\\s+'.$logevent,
        good   => "\$logevent $logevent",
        keep   => "first",
        add    => "last"
        );

  $cpuf and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$cpumultt\\s+.*',
        goodre => '^\\$cpumultt\\s+'.$cpuf,
        good   => "\$cpumultt $cpuf",
        keep   => "first",
        add    => "last"
        );

  $wallf and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$wallmultt\\s+.*',
        goodre => '^\\$wallmultt\\s+'.$wallf,
        good   => "\$wallmultt $wallf",
        keep   => "first",
        add    => "last"
        );

  $idealload and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$idealload\\s+.*',
        goodre => '^\\$idealload\\s+'.$idealload,
        good   => "\$idealload $idealload",
        keep   => "first",
        add    => "last"
        );

  $maxload and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$maxload\\s+.*',
        goodre => '^\\$maxload\\s+'.$maxload,
        good   => "\$maxload $maxload",
        keep   => "first",
        add    => "last"
        );

  $prologalarm and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$prologalarm\\s+.*',
        goodre => '^\\$prologalarm\\s+'.$prologalarm,
        good   => "\$prologalarm $prologalarm",
        keep   => "first",
        add    => "last"
        );

  $nodecheckscript and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$nodecheckscript\\s+.*',
        goodre => '^\\$nodecheckscript\\s+'.$nodecheckscript,
        good   => "\$nodecheckscript $nodecheckscript",
        keep   => "first",
        add    => "last"
        );

  $nodecheckinterval and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$nodecheckinterval\\s+.*',
        goodre => '^\\$nodecheckinterval\\s+'.$nodecheckinterval,
        good   => "\$nodecheckinterval $nodecheckinterval",
        keep   => "first",
        add    => "last"
        );

  if ( $behaviour eq "Torque" ) { # add a $pbsservername line
    # assume first master is the real one
    my $master=(split(/\s+/,$masterlist))[0];
    $master and $changes+=NCM::Check::lines($pbsmomconf,
	backup => ".old",
        linere => '^\\$pbsmastername\\s+.*',
        goodre => '^\\$pbsmastername\\s+'.$master,
        good   => "\$pbsmastername $master",
        keep   => "first",
        add    => "first"
        );
  } 

    my $srvfile="$pbsdir/server_name";
    -e "$srvfile" or ( open SNM,">$srvfile" and close SNM );
    -f "$srvfile" or $self->Error("$srvfile exists but is not a file");

    LC::Check::status("$srvfile",
	owner => "root",
	group => "root",
        mode  => 0644 );

    my $master=(split(/\s+/,$masterlist))[0];
    $master and do { 
      NCM::Check::lines($srvfile,
	backup => ".old",
        linere => '.*',
        goodre => $master,
        good   => $master,
        keep   => "first",
        add    => "first"
        );
      $self->info("Updated file $srvfile with mastername $master");
    };
  

  # create the line(s) with the $usecp direcxties from path hash
  $cnflines="";
  $cnfre="";
  $cnfgoodre="";
  foreach ( keys %usecp ) { 
    $cnflines.='$usecp '.$_." ".$usecp{$_}."\n";
    $cnfgoodre.='\\$usecp '.$_." ".$usecp{$_}."\n";
    $cnfre.='\\$usecp\\s+.*'."\n";
  }
  chomp($cnflines);
  chomp($cnfre);
  chomp($cnfgoodre);
  # have to do all kinds of funcky things to escape the regexs
  $cnfgoodre=~s+/+\\/+g;
  $cnfgoodre=~s+\*+\\\*+g;
  $cnfgoodre=~s+\.+\\\.+g;

  $changes+=NCM::Check::lines($pbsmomconf,
        backup => ".old",
	linere => $cnfre,
	goodre => $cnfgoodre,
	good   => $cnflines,
        keep   => "last",
        add    => "last"
        );


  $self->info("Resources will not be set since they cannot be deleted, sorry");

  $self->debug(1,"Accumulated $changes changes");

  # restart PBS if it is already running AND the config changed
  system("$pbsinitscript status > /dev/null || $pbsinitscript restart") 
    if ($changes);

  return;
}


##########################################################################
sub Unconfigure {
##########################################################################
  my ($self,$config)=@_;

  $self->info("Unconfiguring PBS");
  
  &initpaths($config);

  return;
}


1; #required for Perl modules
